/** \brief Parsing tools for the command line interface.
 *
 * The file contains a set of tools for parsing arguments presented to the command line.
 *
 * \author Michael Gill
 * \date 2 Jan 2020
 */

#ifndef CLPARSER_HPP
#define CLPARSER_HPP

#include <map>

typedef std::map<std::string, std::string> argmap;

/**
 * \brief Parse command line arguments.
 *
 * Parse the arguments according to the usage string @see main.cpp
 *
 * @param argc Size of the C-type array argv
 *
 * @param argv C-type type array containing null teminated strings, command
 * line arguments are \n separated.
 *
 * @param parsed A map @see std::map of arguments followed by specified
 * captures, if any.
 *
 * @return True if the arguments were successfully parsed, false otherwise.
 *
 * @note The parsed parameter is undefined if the return value is false.
 */
bool parseArgs(int argc, char** argv, argmap& parsed);

#endif // CLPARSER_HPP
