#ifndef FARRAY_HPP
#define FARRAY_HPP

#include <vector>

template <typename T>
class FArray {
private:
  T* m_array;
  unsigned int m_size;

public:
  FArray();
  FArray(unsigned int n);
  FArray(const FArray<T>& other);
  ~FArray();

  FArray<T>& operator=(const FArray<T>& other);
  FArray<T>& operator=(const std::vector<T>& other);

  T& operator[](unsigned int index);
  const T& operator[](unsigned int index) const;
  unsigned int size() const;
  void trim(unsigned int size);
};

template <typename T>
FArray<T>::FArray() {
  m_size = 0;
  m_array = new T[2];
}
template <typename T>
FArray<T>::FArray(unsigned int n) {
  m_size = n;
  m_array = new T[n];
}

template <typename T>
FArray<T>::FArray(const FArray<T>& other) :
  m_size(other.m_size) {
  m_array = new T[m_size];

  for (unsigned int i = 0; i < m_size; i++) {
    m_array[i] = other.m_array[i];
  }
}

template <typename T>
FArray<T>::~FArray() {
  delete[] m_array;
}

template <typename T>
FArray<T>& FArray<T>::operator=(const FArray<T>& other) {
  if (this != &other) {
    T* nArray = new T[other.m_size];
    for (unsigned int i = 0; i < other.m_size; i++) {
      nArray[i] = other.m_array[i];
    }

    delete[] m_array;

    m_array = nArray;
    m_size = other.m_size;
  }

  return *this;
}

template <typename T>
FArray<T>& FArray<T>::operator=(const std::vector<T>& other) {
  T* nArray = new T[other.size()];

  for (unsigned int i = 0; i < other.size(); i++) {
    nArray[i] = other[i];
  }

  delete[] m_array;

  m_array = nArray;
  m_size = other.size();
  return *this;
}


template <typename T>
T& FArray<T>::operator[](unsigned int index) {
  return m_array[index];
}

template <typename T>
const T& FArray<T>::operator[](unsigned int index) const {
  return m_array[index];
}

template <typename T>
unsigned int FArray<T>::size() const {
  return m_size;
}

template <typename T>
void FArray<T>::trim(unsigned int size) {
  m_size = size;
}

#endif
