#ifndef UTIL_HPP
#define UTIL_HPP

#include <array>
#include <algorithm>
#include <sstream>
#include <string>
#include <vector>

#define SQUARE_SIZE 10

#define MAX_BLOCKS 6

namespace util {
  std::string join(const std::vector<unsigned int>& list, std::string delim);

  template <typename T, unsigned int I>
  std::string join(const std::array<T, I>& list, std::string delim) {
    std::stringstream ss("");
    for (unsigned int count = 0; count < I - 1; count++) {
      ss << list[count] << delim;
    }
    ss << list[list.size()-1];

    return ss.str();
  }

  template <typename ForwardIt>
  std::string join(ForwardIt first, ForwardIt last, std::string delim) {
    std::stringstream ss("");
    ForwardIt current = first;
    while (current+1 != last) {
      ss << *current << delim;
      current++;
    }
    ss << *current;

    return ss.str();
  }

  template <typename K>
  class citer {
    private:
        std::vector<K> source;
        std::vector<int> indices;
        int move(int k, int last);
    public:
        citer(const std::vector<K>& data, int k);
        std::vector<K> next();
        bool has_next();
  };

  template <class Constainer>
  bool nextPermuation(Constainer& list);
};

template <typename K>
util::citer<K>::citer(const std::vector<K>& data, int k) : source(std::move(data)), indices(k) {
  for (int i = 0; i < k; i++ ) {
      indices[i] = i;
  }
}

template <typename K>
bool util::citer<K>::has_next() {
  return indices[0] >= 0;
}

template <typename K>
std::vector<K> util::citer<K>::next() {
  std::vector<K> ret;

  for (auto it = indices.begin(); it != indices.end(); it++) {
      ret.push_back(source[*it]);
  }

  int k = indices.size() - 1;

  move(k, source.size() - 1);

  return ret;
}

template <typename K>
int util::citer<K>::move(int k, int last) {
  if (indices[k] < last) {
      indices[k]++;
  } else {
      if (k <= 0) {
          indices[0] = -1;
          return 0;
      }
      int moved = move(k-1, last-1);
      indices[k] = moved + 1;
  }

  return indices[k];
}

#include <iostream>

template <typename Container>
bool util::nextPermuation(Container& list) {
  unsigned int index = list.size()-1;
  while (index > 0 && list[index] < list[index-1]) index--;

  if (index == 0) {
    return false;
  }

  unsigned int pivot = index-1;
  index = list.size()-1;
  while (list[index] < list[pivot]) index--;

  std::swap(list[index], list[pivot]);
  std::reverse(list.begin()+pivot+1, list.end());

  return true;
}

#endif // UTIL_HPP
