#include "diagonals.hpp"

#include <bitset>
#include <numeric>

#include <assert.h>

BitwiseDiag::BitwiseDiag() : low(0), vals() {
}

unsigned int BitwiseDiag::get(unsigned int index) const {
  // if (index < 6) {
  //   uint_fast64_t mask = 1;
  //   mask = ((mask << 10) - 1) << (index*10);
  //
  //   uint_fast64_t item = (low & mask) >> (index*10);
  //   return popCount(item-1);
  // } else if (index < 10) {
  //   unsigned int index_tmp = index - 6;
  //
  //   uint_fast64_t mask = 1;
  //   mask = ((mask << 10) - 1) << (index_tmp*10);
  //
  //   uint_fast64_t item = (high & mask) >> (index_tmp*10);
  //   return popCount(item-1);
  // }

  return vals[index];
  // return 0;
}

void BitwiseDiag::set(unsigned int row, unsigned int col) {
  // We are going to perform assignment to low. First we need to bitmask the
  // section of low that is being accesses back to 0. To do this we take a set
  //of 0's
  if (row < 10) {
    __uint128_t mask = 1;
    mask = (mask << 100) - 1;

    __uint128_t cut = 1;
    cut = cut << 10;
    cut--;
    cut = cut << (row*10);

    mask = mask ^ cut;

    low = mask & low;


    __uint128_t bit = 1;
    bit = (bit << (col)) << (row*10);

    low = bit | low;
  }

  vals[row] = col;
}

bool BitwiseDiag::conflict(const BitwiseDiag& d1) const {
  return (d1.low & low) > 0;
}

bool BitwiseDiag::conflict(const BitwiseDiag *d) const {
  return (d->low & low) > 0;
}

bool BitwiseDiag::conflict(const __uint128_t& cell) const {
  return (low & cell) > 0;
}

bool BitwiseDiag::intersect(const BitwiseDiag& d1) const {
  __uint128_t x = low & d1.low;
  if (x == 0) {
    return false;
  }
  return (x & (x-1)) == 0;
}

bool BitwiseDiag::intersect(const BitwiseDiag *d) const {
  __uint128_t x = low & d->low;
  if (x == 0) {
    return false;
  }
  return (x & (x-1)) == 0;
}

SimpleDiag::SimpleDiag() : diagonal() {}

unsigned int SimpleDiag::get(unsigned int i) const {
  return diagonal[i];
}

void SimpleDiag::set(unsigned int i, unsigned int val) {
  diagonal[i] = val;
}

bool SimpleDiag::conflict(const SimpleDiag& d) {
  for (unsigned int i = 0; i < diagonal.size(); i++) {
    if (diagonal[i] == d.diagonal[i])
      return true;
  }

  return false;
}

bool SimpleDiag::intersect(const SimpleDiag& d) {
  unsigned int count = 0;
  for (unsigned int i = 0; i < diagonal.size(); i++) {
    if (diagonal[i] == d.diagonal[i]) {
      if (count == 1)
        return false;

      count++;
    }
  }

  return count == 1;
}

void SimpleDiag::show() {}

template <typename T>
struct incrementer {
  T current;
  incrementer(T start) : current(start) {}
  T operator()() { return current++; }
};

void BitwiseDiag::print(std::ostream& stream) const {
  for (unsigned int idx = 0; idx < N; idx++) {
    stream << get(idx) << " ";
  }
}

void SimpleDiag::print(std::ostream& stream) const {
  for (unsigned int idx = 0; idx < N; idx++) {
    stream << get(idx) << " ";
  }
}

std::ostream& operator << (std::ostream& stream, const Diag& diag) {
  diag.print(stream);
  return stream;
}

std::ostream& operator << (std::ostream& stream,
  const std::vector<Diag>& diags) {
  if (diags.size() > N) {
    stream << "ERROR!!!! Bigger than a full square!\n";
  }

  Square square;
  for (unsigned int i = 0; i < N; i++) {
    for (unsigned int j = 0; j < N; j++) {
      square[i][j] = 0;
    }
  }

  for (unsigned int idx = 0; idx < diags.size(); idx++) {
    for (unsigned int row = 0; row < N; row++) {
      square[row][diags[idx].get(row)] = idx+1;
    }
  }

  stream << square;
  return stream;
}

std::ostream& operator << (std::ostream& stream,
  const std::vector<Diag*>& diags) {
  if (diags.size() > N) {
    stream << "ERROR!!!! Bigger than a full square!\n";
  }

  Square square;
  for (unsigned int i = 0; i < N; i++) {
    for (unsigned int j = 0; j < N; j++) {
      square[i][j] = 0;
    }
  }

  for (unsigned int idx = 0; idx < diags.size(); idx++) {
    for (unsigned int row = 0; row < N; row++) {
      square[row][diags[idx]->get(row)] = idx+1;
    }
  }

  stream << square;
  return stream;
}

// FSTACKS
std::ostream& operator << (std::ostream& stream,
  const FStack<Diag>& diags) {
  if (diags.size() > N) {
    stream << "ERROR!!!! Bigger than a full square!\n";
  }

  Square square;
  for (unsigned int i = 0; i < N; i++) {
    for (unsigned int j = 0; j < N; j++) {
      square[i][j] = 0;
    }
  }

  for (unsigned int idx = 0; idx < diags.size(); idx++) {
    for (unsigned int row = 0; row < N; row++) {
      square[row][diags[idx].get(row)] = idx+1;
    }
  }

  stream << square;
  return stream;
}

std::ostream& operator << (std::ostream& stream,
  const FStack<Diag*>& diags) {
  if (diags.size() > N) {
    stream << "ERROR!!!! Bigger than a full square!\n";
  }

  Square square;
  for (unsigned int i = 0; i < N; i++) {
    for (unsigned int j = 0; j < N; j++) {
      square[i][j] = 0;
    }
  }

  for (unsigned int idx = 0; idx < diags.size(); idx++) {
    for (unsigned int row = 0; row < N; row++) {
      square[row][diags[idx]->get(row)] = idx+1;
    }
  }

  stream << square;
  return stream;
}

bool DiagGenerator::hasFrequencySpace(uint_mod<2> square, unsigned int row,
  unsigned int col) {
  for (unsigned int idx = 0; idx < m_freqHits.size(); idx++) {
    if (m_squares[idx+2][row][col]) {
      if (m_freqHits[idx] >= m_freqCounts[idx]) {
        return false;
      }
    }
  }

  if (m_squares[(unsigned int)(square+1)][row][col]) {
    if (m_oHits >= m_oCounts[(unsigned int)(square+1)]) {
      return false;
    }
  }

  return true;
}

void DiagGenerator::updateFrequencies(uint_mod<2> square, unsigned int row,
  unsigned int col) {
  for (unsigned int idx = 0; idx < m_freqHits.size(); idx++) {
    if (m_squares[idx+2][row][col]) {
      m_freqHits[idx]++;
    }
  }

  if (m_squares[(unsigned int)(square+1)][row][col]) {
    m_oHits++;
  }
}

void DiagGenerator::restoreFrequencies(uint_mod<2> square, unsigned int row,
  unsigned int col) {
  for (unsigned int idx = 0; idx < m_freqHits.size(); idx++) {
    if (m_squares[idx+2][row][col]) {
      m_freqHits[idx]--;
    }
  }

  if (m_squares[(unsigned int)(square+1)][row][col]) {
    m_oHits--;
  }
}

bool DiagGenerator::symbolMatch(unsigned int square, unsigned int row, unsigned int col) {
  return m_squares[square][row][col] == m_squares[square][0][m_diag.get(0)];
}

void DiagGenerator::generate(const uint_mod<2> square,
  const std::vector<unsigned int>& available) {
  if (available.size() == 0) {
    for (unsigned int idx = 0; idx < m_freqCounts.size(); idx++) {
      if (m_freqCounts[idx] != m_freqHits[idx]) {
        return;
      }
    }
    if (m_oHits != m_oCounts[(unsigned int)(square+1)]) {
      return;
    }
    if (symbolMatch((unsigned int)square, N-1, m_diag.get(N-1))) {
      m_diagonals[(unsigned int)square].push_back(m_diag);
    }
  }

  unsigned int row = N - available.size();

  for (unsigned int col : available) {
    if (row == 0 || symbolMatch((unsigned int)square, row, col)) {
      if (hasFrequencySpace(square, row, col)) {

        m_diag.set(row, col);

        updateFrequencies(square, row, col);

        std::vector<unsigned int> newAvailable;
        for (unsigned int val : available) {
          if (val != col) {
              newAvailable.push_back(val);
          }
        }

        generate(square, newAvailable);

        restoreFrequencies(square, row, col);
      }
    }
  }
}

DiagGenerator::DiagGenerator(const std::vector<Square>& squares ) :
m_squares(squares), m_freqCounts(squares.size()-2, 0), m_oCounts({0,0}) {

  // Calculate the frequency count
  for (unsigned int sq = 2; sq < squares.size(); sq++) {
    for (unsigned int idx = 0; idx < N; idx++) {
      if (m_squares[sq][0][idx]) {
        m_freqCounts[sq-2]++;
      }
    }
  }

  // Find the s and t counts. This is because they act as frequency squares in
  // the unresolved case.
  for (unsigned int square = 0; square < 2; square++) {
    for (unsigned int idx = 0; idx < N; idx++) {
      if (m_squares[square][0][idx]) {
        m_oCounts[square]++;
      }
    }
  }
}

void DiagGenerator::generate() {
  std::vector<unsigned int> available(N);
  std::iota(available.begin(), available.end(), 0);

  m_oHits = 0;
  m_freqHits = std::vector<unsigned int>(m_squares.size()-2, 0);
  generate(0, available);

  m_oHits = 0;
  m_freqHits = std::vector<unsigned int>(m_squares.size()-2, 0);
  generate(1, available);
}

std::vector<Diag>& DiagGenerator::getSDiagonals() {
  return m_diagonals[0];
}

std::vector<Diag>& DiagGenerator::getTDiagonals() {
  return m_diagonals[1];
}

void DiagGenerator::print(std::ostream& stream) const {
  stream << "Squares:\n" << m_squares << "\n";

  stream << "Frequency Counts: \n" << util::join(m_freqCounts, " ") << "\n";

  stream << "oCounts: \n" << util::join<unsigned int, 2>(m_oCounts, " ") << "\n";

  stream << "freqHits: \n" << util::join(m_freqHits, " ") << "\n";

  stream << "oHits: \n" << m_oHits << "\n";
}

std::ostream& operator << (std::ostream& stream, const DiagGenerator& gen) {
  gen.print(stream);

  return stream;
}
