#include "solver.hpp"

// #define MESSAGELOG
#include "logger.hpp"

//#define BLOCKLOG

// ############################### Start DTrickSolver ##########################

unsigned int squareCount;
FArray<__uint128_t> cells;

#ifdef BLOCKLOG
unsigned int depth;
#endif

unsigned int solve(
  std::ostream& stream,
  const FArray<FArray<__uint128_t>>& diagonals
) {
  // Setup cells and active cells.
  cells = FArray<__uint128_t>(N*N);
  FArray<FArray<unsigned int>> availableCells(2);
  for (unsigned int s = 0; s < 2; s++)
  {
    availableCells[s] = FArray<unsigned int >(N*N);
  }

  for (unsigned int count = 0; count < N*N; count++) {
    __uint128_t bit = 1;
    cells[count] = (bit << count);
    // TODO: Make this more efficient
    availableCells[0][count] = count;
    availableCells[1][count] = count;
  }

  FArray<FStack<__uint128_t>> partial(2);

  for (unsigned int count = 0; count < 2; count++) {
    partial[count] = FStack<__uint128_t>(10);
  }

  #ifdef BLOCKLOG
  depth = 0;
  #endif

  squareCount = 0;
  solve(stream, diagonals, availableCells, partial);

  return squareCount;
}

bool conflict(__uint128_t a, __uint128_t b) {
  return (a & b) > 0;
}

bool intersect(__uint128_t a, __uint128_t b) {
  __uint128_t x = a & b;
  if (x == 0) {
    return false;
  }
  return (x & (x - 1)) == 0;
}

unsigned int countDiagHit(
  const __uint128_t cell,
  const FArray<__uint128_t>& diagonals,
  const unsigned int ceil,
  unsigned int dSize
) {
  unsigned int count = 0;
  for (unsigned int index = 0; index < dSize; index++) {
    if (conflict(diagonals[index], cell)) {
      count ++;
      if (count > ceil) {
        return count;
      }
    }
  }

  return count;
}

unsigned int countTransHit(
  const __uint128_t cell,
  const FArray<__uint128_t>& transversals,
  const unsigned int ceil,
  unsigned int dSize
) {
  unsigned int count = 0;
  for (unsigned int index = 0; index < dSize; index++) {
    if (conflict(transversals[index], cell)) {
      count ++;
      if (count > ceil) {
        return count;
      }
    }
  }

  return count;
}

// Find the postiion of the lowest incident cell relative cells array.
unsigned int findMinCellTrans(
  const FArray<FArray<__uint128_t>>& diagonals,
  const FArray<unsigned int>& availableCells,
  const unsigned int square,
  unsigned int& count
) {
  unsigned int pos = availableCells[0];
  unsigned int dSize = diagonals[square].size();
  count = countTransHit(cells[pos], diagonals[(square+1)%2],
    diagonals[(square+1)%2].size(), dSize);

  for (unsigned int index = 1; index < availableCells.size(); index++) {
    unsigned int tmp = countDiagHit(cells[availableCells[index]],
      diagonals[(square+1)%2], count, dSize);

    if (tmp < count) {
      count = tmp;
      pos = availableCells[index];
    }
  }

  return pos;
}

// Find the postiion of the lowest incident cell relative cells array.
unsigned int findMinCell(
  const FArray<FArray<__uint128_t>>& diagonals,
  const FArray<unsigned int>& availableCells,
  const unsigned int square,
  unsigned int& count
) {
  unsigned int pos = availableCells[0];
  unsigned int dSize = diagonals[square].size();
  count = countDiagHit(cells[pos], diagonals[square],
    diagonals[square].size(), dSize);

  for (unsigned int index = 1; index < availableCells.size(); index++) {
    unsigned int tmp = countDiagHit(cells[availableCells[index]],
      diagonals[square], count, dSize);

    if (tmp < count) {
      count = tmp;
      pos = availableCells[index];
    }
  }

  return pos;
}

void trackCells(
  const __uint128_t diag,
  const FArray<FArray<unsigned int>>& availableCells,
  unsigned int square,
  unsigned int altSquare,
  FArray<FArray<unsigned int>>& nAvailCells
) {
  nAvailCells[altSquare] = availableCells[altSquare];
  nAvailCells[square] = FArray<unsigned int>(availableCells[square].size());
  unsigned int cellCount = 0;
  for (unsigned int cIdx = 0; cIdx < availableCells[square].size(); cIdx++) {
    if (!conflict(diag, cells[availableCells[square][cIdx]])) {
      nAvailCells[square][cellCount++] = availableCells[square][cIdx];
    }
  }
  nAvailCells[square].trim(cellCount);
}

void trackDiagonals(
  const __uint128_t diag,
  const FArray<FArray<__uint128_t>>& diagonals,
  unsigned int square,
  unsigned int altSquare,
  FArray<FArray<__uint128_t>>& nDiagonals
) {
  nDiagonals[square] = FArray<__uint128_t>(diagonals[square].size());
  nDiagonals[altSquare] = FArray<__uint128_t>(diagonals[altSquare].size());
  unsigned int dCount = 0;
  unsigned int dSize = diagonals[square].size();
  for (unsigned int dIdx = 0; dIdx < dSize; dIdx++) {
    if (!conflict(diag, diagonals[square][dIdx])) {
      nDiagonals[square][dCount++] = diagonals[square][dIdx];
    }
  }
  nDiagonals[square].trim(dCount);

  // Now for the alt square
  dCount = 0;
  dSize = diagonals[altSquare].size();
  for (unsigned int dIdx = 0; dIdx < dSize; dIdx++) {
    if (intersect(diag, diagonals[altSquare][dIdx])) {
      nDiagonals[altSquare][dCount++] = diagonals[altSquare][dIdx];
    }
  }
  nDiagonals[altSquare].trim(dCount);
}

void trackCellDiagonals(
  __uint128_t cell,
  const FArray<__uint128_t>& diagonals,
  FArray<__uint128_t>& cellDiagonals
) {
  unsigned int number = 0;
  unsigned int dSize = cellDiagonals.size();
  for (unsigned int index = 0; index < dSize; index++) {
    if (conflict(diagonals[index], cell)) {
      cellDiagonals[number++] = diagonals[index];
      if (number == dSize) {
        break;
      }
    }
  }
}

void solve(
  std::ostream& stream,
  const FArray<FArray<__uint128_t>>& diagonals,
  const FArray<FArray<unsigned int>>& availableCells,
  FArray<FStack<__uint128_t>>& partial
) {

  // This is for profiling.
  // if (squareCount  > 100) {
  //   return;
  // }

  // Base case: Success
  if (partial[0].size() == N && partial[1].size() == N) {
    stream << squareCount++ << "\n";
    for (unsigned int s = 0; s < 2; s++) {
      stream << partial[s];
    }

    stream << "\n";

    return;
  }

  // Detemine what square we care about.
  unsigned int square = 0;
  unsigned int altSquare = 1;
  if (partial[0].size() > partial[1].size()) {
    square = 1;
    altSquare = 0;
  }


  // Decide the cell and the number of intersecting diagonals.
  unsigned int count;
  unsigned int pos = findMinCell(diagonals, availableCells[square],
    square, count);

  #ifdef BLOCKLOG
  std::cerr << std::endl << depth << " : " << count << " : " << pos << std::endl;
  if (count == 0) {
    for (unsigned int i = 0; i < 10; i ++) {
      for (unsigned int j = 0; j < 10; j++) {
        std::cerr << countDiagHit(cells[i*10+j],
          diagonals[square], diagonals[square].size(), diagonals[square].size()) << "\t";
      }
      std::cerr << "\n";
    }
  }
  #endif

  // Base case: Fail 2
  if (count == 0) {
    return;
  }

  // Determine all the possible diagonals.
  // FArray<Diag*> cellDiagonals(count);
  // trackCellDiagonals(cells[pos], diagonals[square], cellDiagonals);

  // for each diagonal related to the cell. Lets fix things.
  // for (unsigned int index = 0; index < cellDiagonals.size(); index++) {
  //
  //   Diag* diag = cellDiagonals[index];

  unsigned int used = 0;
  for (unsigned int index = 0; index < diagonals[square].size() && used < count; index++ ) {

    __uint128_t diag = diagonals[square][index];
    if (!conflict(diag, cells[pos])) {
      continue;
    }

    used++;

    // Switch off all the unwanted cells. and construct a new cell list.
    FArray<FArray<unsigned int>> nAvailCells(2);
    trackCells(diag, availableCells, square, altSquare, nAvailCells);

    // Construct a new diagonal list for bith the current square and the alt.
    FArray<FArray<__uint128_t>> nDiagonals(2);
    trackDiagonals(diag, diagonals, square, altSquare, nDiagonals);

    partial[square].push(diag);

    #ifdef BLOCKLOG
    depth ++;
    #endif

    solve(stream, nDiagonals, nAvailCells, partial);

    #ifdef BLOCKLOG
    depth --;
    #endif

    partial[square].pop();
  }
}

FArray<unsigned int> unpackDiag(__uint128_t diag) {
  FArray<unsigned int> ret(N);
  for (unsigned int index = 0; index < N; index++) {
    __uint128_t mask = 1;
    mask = ((mask << N) - 1) << (index*N);

    __uint128_t item = (diag & mask) >> (index*N);
    ret[index] = popCount(item-1);
  }

  return ret;
}

std::ostream& operator<<(
  std::ostream& stream, const FStack<__uint128_t>& list
) {
  Square square;
  for (unsigned int symbol = 0 ; symbol < N; symbol++) {
    FArray<unsigned int> unpacked = unpackDiag(list[symbol]);
    for (unsigned int index = 0; index < N; index++) {
      square[index][unpacked[index]] = symbol+1;
    }
  }

  stream << square;

  return stream;
}

// ############################### End DTrickSolver ###############################
