// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// Handles input of LS from istream. Decoder input specifications can be
// found in the manual.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LATIN10_DIAGONALS_DIAGONALS_HPP_
#define LATIN10_DIAGONALS_DIAGONALS_HPP_

#include <algorithm>
#include <array>
#include <iostream>
#include <vector>

#include "definitions.hpp"
#include "number/integer.hpp"
//#include "util/fstack.hpp"
//#include "util/list.hpp"

// Diag positions are 0 indexed
struct BitwiseDiag {
  __uint128_t low;
  std::array<unsigned int, N> vals;

  BitwiseDiag();
  BitwiseDiag(std::string str);

  unsigned int get(unsigned int index) const;
  void set(unsigned int i, unsigned int val);
  bool conflict(const BitwiseDiag &d) const;
  bool conflict(const BitwiseDiag *d) const;
  bool conflict(const __uint128_t &cell) const;
  bool intersect(const BitwiseDiag &d) const;
  bool intersect(const BitwiseDiag *d) const;
  void show();
};

struct SimpleDiag {
  std::array<unsigned int, N> diagonal;
  bool active;

  SimpleDiag();

  unsigned int get(unsigned int index) const;
  void set(unsigned int i, unsigned int val);
  bool conflict(const SimpleDiag &d);
  bool intersect(const SimpleDiag &d);
  void show();
};

// typedef SimpleDiag Diag;
typedef BitwiseDiag Diag;

std::ostream &operator<<(std::ostream &stream, const Diag &diag);

std::ostream &operator<<(std::ostream &stream, const std::vector<Diag> &diags);

std::ostream &operator<<(std::ostream &stream,
                         const std::vector<Diag *> &diags);

//std::ostream &operator<<(std::ostream &stream, const FStack<Diag *> &diags);

std::vector<std::vector<Diag>>
gen_diagonals(const std::vector<Square> &squares);

/**
 * \brief Generates pairs of diagonalsthat constitute possible solutions to the
 * MOLS satisfying the frequency squares.
 *
 * @param mols A set of partially completed mols to be completed
 * @param fSquares A set of frequency squares
 * @param diagonalPairs Contains the restults of attemting to complete the
 * diagonals.
 *
 */
// void generateSquareDiagonals(const uint_mod<2> type, const MOLSPair& mols,
// const FrequencySquares& fSquares, std::vector<Diag>& diagonals);

#endif // LATIN10_DIAGONALS_DIAGONALS_HPP_
