// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// Handles ouput of squares using formats detailed in io section of manual.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LATIN10_IO_ENCODER_HPP_
#define LATIN10_IO_ENCODER_HPP_

#include <iostream>
#include <vector>
#include <algorithm>

#include "definitions.hpp"
#include "number/integer.hpp"

// Base struct contract for encoder subclasses. The LS contained in square is
// witten to the output stream.
struct Encoder {
  virtual void encode(std::ostream& stream, const Square& square) = 0;
  virtual ~Encoder();
};

// See manual for details on the numeric encoding.
struct NumericEncoder : public Encoder {
  virtual void encode(std::ostream& stream, const Square& square);
};

// See manual for details on the alphabetical encoding.
struct AlphaEncoder : public Encoder {
  virtual void encode(std::ostream& stream, const Square& square);
};

// See manual for details on the bit packed encoding.
struct BitEncoder : public Encoder {
  virtual void encode(std::ostream& stream, const Square& square);
};

// See manual for details on the factorial encoding.
struct FactorialEncoder : public Encoder {
  virtual void encode(std::ostream& stream, const Square& square);
};

std::string frequencyPacking(const Square& square);
#endif // LATIN10_IO_ENCODER_HPP_
