// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// A set of program wide definitions and typename declarations. All header files
// should include this file to be ensure project consistency.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LATIN10_TEMPLATE_ROW_HPP_
#define LATIN10_TEMPLATE_ROW_HPP_

#include <array>
#include <numeric>

#include "definitions.hpp"
#include "util/citer.hpp"
#include "number/integer.hpp"

// Bitwise operations to deal with setting a row and clearing a row.
#define SET_ROW_SYMBOL(ROW, I, S) ((ROW) | ((uint_fast64_t)1 << ((S)*(N)+(I))))
#define CLEAR_ROW_SYMBOL(ROW, I, S) ((ROW) & (~((uint_fast64_t)1 << ((S)*(N)+(I)))))
#define GET_ROW_SYM(ROW, I, S) ((ROW) & ((uint_fast64_t)1 << ((S)*(N)+(I))))

typedef uint_fast64_t TRow;

void generateRow(
  std::vector<TRow>& rows,
  const std::vector<size_t>& gSizes,
  const std::array<size_t, GDD_B_SIZES>& bSizes,
  const bool inRelation,
  const size_t& colRelation
);

void printTRow(const TRow&);

#endif
