// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// A set of program wide definitions and typename declarations. All header files
// should include this file to be ensure project consistency.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LATIN10_TEMPLATE_TEMPLATE_HPP_
#define LATIN10_TEMPLATE_TEMPLATE_HPP_

#include <iostream>

#include "definitions.hpp"
#include "diagonals/diagonals.hpp"
#include "number/integer.hpp"

class Template {
private:
  // Weight of the template
  size_t weight;
  // Contains the template data
  std::array<std::vector<size_t>, N * N> data;

  // Contains a list of host start diagonals. Empty if none exists.
  bool is_hot = false;
  std::vector<std::array<std::vector<Diag>,2>> hs_list;

public:
  Template(size_t weight);

  // Include a symbol in the relation
  void set(size_t row, size_t col, size_t sym);

  // Remove symbol from relation
  void clear(size_t row, size_t col, size_t sym);

  // Returns true if the cell (row, col) contains the symbol sym.
  bool get(size_t row, size_t col, size_t sym) const;

  // Get the weight of the template
  size_t getWeight() const;

  // Swap the order of the squares in the template
  void swapSym(size_t s1, size_t s2);

  // Convert the template to a set of squares.
  std::vector<Square> toSquares() const;

  void hs_reserve(size_t orbits);

  void hs_orbit_reserve(size_t count, size_t square, size_t orbit);

  void add_hs_diag(Diag d, size_t square, size_t orbit);

  std::vector<std::array<std::vector<uint128_t>, 2>> get_hs_diag_as_uint128_t();

  // Friend linking to output operator
  friend std::ostream &operator<<(std::ostream &os, const Template &tmpl);
};

#endif
