#include "diagonals/diagonals.hpp"

#include <bitset>
#include <numeric>

#include <assert.h>

BitwiseDiag::BitwiseDiag() : low(0), vals() {}

BitwiseDiag::BitwiseDiag(std::string str) : low(0), vals() {
  std::array<bool, N> seen;
  for (auto &val : seen) {
    val = false;
  }

  for (size_t i = 0; i < N - 1; i++) {
    size_t j = hexctoi(str[i]);
    set(i, j);

    seen[j] = true;
  }

  size_t i = 0;
  while (seen[i]) {
    i++;
  }
  set(N - 1, i);
}

unsigned int BitwiseDiag::get(unsigned int index) const {
  // if (index < 6) {
  //   uint_fast64_t mask = 1;
  //   mask = ((mask << 10) - 1) << (index*10);
  //
  //   uint_fast64_t item = (low & mask) >> (index*10);
  //   return popCount(item-1);
  // } else if (index < 10) {
  //   unsigned int index_tmp = index - 6;
  //
  //   uint_fast64_t mask = 1;
  //   mask = ((mask << 10) - 1) << (index_tmp*10);
  //
  //   uint_fast64_t item = (high & mask) >> (index_tmp*10);
  //   return popCount(item-1);
  // }

  return vals[index];
  // return 0;
}

void BitwiseDiag::set(unsigned int row, unsigned int col) {
  // We are going to perform assignment to low. First we need to bitmask the
  // section of low that is being accesses back to 0. To do this we take a set
  // of 0's
  if (row < 10) {
    __uint128_t mask = 1;
    mask = (mask << 100) - 1;

    __uint128_t cut = 1;
    cut = cut << 10;
    cut--;
    cut = cut << (row * 10);

    mask = mask ^ cut;

    low = mask & low;

    __uint128_t bit = 1;
    bit = (bit << (col)) << (row * 10);

    low = bit | low;
  }

  vals[row] = col;
}

bool BitwiseDiag::conflict(const BitwiseDiag &d1) const {
  return (d1.low & low) > 0;
}

bool BitwiseDiag::conflict(const BitwiseDiag *d) const {
  return (d->low & low) > 0;
}

bool BitwiseDiag::conflict(const __uint128_t &cell) const {
  return (low & cell) > 0;
}

bool BitwiseDiag::intersect(const BitwiseDiag &d1) const {
  __uint128_t x = low & d1.low;
  if (x == 0) {
    return false;
  }
  return (x & (x - 1)) == 0;
}

bool BitwiseDiag::intersect(const BitwiseDiag *d) const {
  __uint128_t x = low & d->low;
  if (x == 0) {
    return false;
  }
  return (x & (x - 1)) == 0;
}

SimpleDiag::SimpleDiag() : diagonal() {}

unsigned int SimpleDiag::get(unsigned int i) const { return diagonal[i]; }

void SimpleDiag::set(unsigned int i, unsigned int val) { diagonal[i] = val; }

bool SimpleDiag::conflict(const SimpleDiag &d) {
  for (unsigned int i = 0; i < diagonal.size(); i++) {
    if (diagonal[i] == d.diagonal[i])
      return true;
  }

  return false;
}

bool SimpleDiag::intersect(const SimpleDiag &d) {
  unsigned int count = 0;
  for (unsigned int i = 0; i < diagonal.size(); i++) {
    if (diagonal[i] == d.diagonal[i]) {
      if (count == 1)
        return false;

      count++;
    }
  }

  return count == 1;
}

void SimpleDiag::show() {}

template <typename T> struct incrementer {
  T current;
  incrementer(T start) : current(start) {}
  T operator()() { return current++; }
};

std::vector<std::vector<size_t>> get_freq(const std::vector<Square> &squares) {
  std::vector<std::vector<size_t>> result(squares.size());

  for (size_t square = 0; square < squares.size(); square++) {
    std::array<size_t, N> counts;
    for (auto &val : counts) {
      val = 0;
    }

    for (size_t row = 0; row < N; row++) {
      counts[squares[square][row * N]]++;
    }

    size_t pos = 0;
    while (counts[pos] != 0) {
      result[square].push_back(counts[pos]);
      pos++;
    }
  }

  return result;
}

struct FrequencyTable {
  size_t freq_symbol = 0;
  std::vector<std::vector<size_t>> freq_counts;
  std::vector<std::vector<size_t>> freq_collisions;
};

void gen_diagonals(const std::vector<Square> &squares,
                   std::vector<Diag> &diagonals, Diag &partial, size_t row,
                   std::array<bool, N> &cols, const size_t layer,
                   FrequencyTable &freq_table) {
  if (row == N) {
    diagonals.push_back(partial);
    return;
  }

  for (size_t col = 0; col < N; col++) {
    if (cols[col]) {
      continue;
    }

    if (row == 0) {
      freq_table.freq_symbol = squares[layer][row * N + col];
    } else if (freq_table.freq_symbol != squares[layer][row * N + col]) {
      continue;
    }

    bool able_to_fit = true;

    for (size_t square = 0; square < squares.size() && able_to_fit; square++) {
      if (square == layer) {
        continue;
      }

      if (freq_table.freq_counts[square][squares[square][row * N + col]] >=
          freq_table.freq_collisions[square][squares[square][row * N + col]]) {
        able_to_fit = false;
      }
    }

    if (!able_to_fit) {
      continue;
    }

    cols[col] = true;

    partial.set(row, col);
    for (size_t square = 0; square < squares.size(); square++) {
      if (square == layer) {
        continue;
      }

      freq_table.freq_counts[square][squares[square][row * N + col]]++;
    }

    gen_diagonals(squares, diagonals, partial, row + 1, cols, layer,
                  freq_table);

    cols[col] = false;

    for (size_t square = 0; square < squares.size(); square++) {
      if (square == layer) {
        continue;
      }

      freq_table.freq_counts[square][squares[square][row * N + col]]--;
    }
  }
}

std::vector<std::vector<Diag>>
gen_diagonals(const std::vector<Square> &squares) {
  std::vector<std::vector<size_t>> freq_collisions = get_freq(squares);

  std::vector<std::vector<size_t>> freq_counts(freq_collisions.size());
  for (size_t i = 0; i < freq_counts.size(); i++) {
    freq_counts[i] = std::vector<size_t>(freq_collisions[i].size(), 0);
  }

  FrequencyTable freq_table;
  freq_table.freq_counts = freq_counts;
  freq_table.freq_collisions = freq_collisions;

  Diag partial;

  std::vector<std::vector<Diag>> diagonals(squares.size());

  std::array<bool, N> cols;
  for (auto &col : cols) {
    col = false;
  }

  for (size_t layer = 0; layer < squares.size(); layer++) {
    gen_diagonals(squares, diagonals[layer], partial, 0, cols, layer,
                  freq_table);
  }

  return diagonals;
}
