// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// A set of program wide definitions and typename declarations. All header files
// should include this file to be ensure project consistency.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "template/template.hpp"

Template::Template(const size_t weight) : weight(weight) {
  for (size_t index = 0; index < N * N; index++) {
    data[index] = std::vector<size_t>(weight, 0);
  }
}

void Template::set(size_t row, size_t col, size_t sym) {
  data[row * N + col][sym] = 1;
}

void Template::clear(size_t row, size_t col, size_t sym) {
  data[row * N + col][sym] = 0;
}

bool Template::get(size_t row, size_t col, size_t sym) const {
  return data[row * N + col][sym];
}

size_t Template::getWeight() const { return weight; }

void Template::swapSym(size_t s1, size_t s2) {
  for (size_t row = 0; row < N; row++) {
    for (size_t col = 0; col < N; col++) {
      std::swap(data[row * N + col][s1], data[row * N + col][s2]);
    }
  }
}

std::vector<Square> Template::toSquares() const {
  std::vector<Square> squares(weight);

  for (size_t row = 0; row < N; row++) {
    for (size_t col = 0; col < N; col++) {
      for (size_t sym = 0; sym < weight; sym++) {
        if (get(row, col, sym)) {
          squares[sym][row * N + col] = 1;
        } else {
          squares[sym][row * N + col] = 0;
        }
      }
    }
  }

  return squares;
}

void Template::hs_reserve(size_t orbits) {
  hs_list.reserve(orbits);
  for (size_t i = 0; i < orbits; i++) {
    std::array<std::vector<Diag>, 2> orbit;
    hs_list.push_back(orbit);
  }
}

void Template::hs_orbit_reserve(size_t count, size_t square, size_t orbit) {
  hs_list[orbit][square].reserve(count);
}

void Template::add_hs_diag(Diag d, size_t square, size_t orbit) {
  hs_list[orbit][square].push_back(d);
}

std::vector<std::array<std::vector<uint128_t>, 2>> Template::get_hs_diag_as_uint128_t() {
  std::vector<std::array<std::vector<uint128_t>, 2>> result;
  result.reserve(hs_list.size());
  for (size_t i = 0; i < hs_list.size(); i++) {
    std::array<std::vector<uint128_t>, 2> orbit;
    orbit[0].reserve(hs_list[i][0].size());
    orbit[1].reserve(hs_list[i][1].size());

    for (Diag &val : hs_list[i][0]) {
      orbit[0].push_back(val.low);
    }

    for (Diag &val : hs_list[i][1]) {
      orbit[1].push_back(val.low);
    }
    result.push_back(orbit);
  }

  return result;
}

std::ostream &operator<<(std::ostream &os, const Template &tmpl) {
  for (size_t row = 0; row < N; row++) {
    for (size_t col = 0; col < N; col++) {
      size_t val = 0;
      for (size_t sym = 0; sym < tmpl.getWeight(); sym++) {
        val += 1 << sym;
      }

      os << itohexc(val);
    }
    if (row < N) {
      os << "\n";
    }
  }

  return os;
}
